
// Filename: fft.cpp

//----------------------------------------------------------------------
//  Copyright (c) 2010 by Doulos.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//----------------------------------------------------------------------

#include "svdpi.h"

#include <math.h>
#include <iostream>
using namespace std;

typedef struct
{
  float real;
  float imag;
} complex;

static int N;
#define PI 3.1415926535

complex* x;
complex* W;

void do_butterfly(int TopVal,int BotVal,int R) {

  complex Temp;

  // Complex multiply
  Temp.real=(x[BotVal].real*W[R].real) - (x[BotVal].imag*W[R].imag);
  Temp.imag=(x[BotVal].real*W[R].imag) + (x[BotVal].imag*W[R].real);

  // Bottom Value Calculation
  x[BotVal].real=x[TopVal].real - Temp.real;
  x[BotVal].imag=x[TopVal].imag - Temp.imag;

  // Top Value Calculation
  x[TopVal].real=x[TopVal].real + Temp.real;
  x[TopVal].imag=x[TopVal].imag + Temp.imag;
}

void do_fft() {
  complex Temp;
  int BSep,BWidth,P,R,TopVal,BotVal;

//  cout << endl
//    << "******************************" << endl
//    << "Bit - reversal reordering data" << endl
//    << "******************************" << endl << endl;

  // Reorder input data (bit-reversal addressing)
  int k;
  int j = 1;
  for (int i = 1; i < N; i++) {
    if (i < j) {
      // Swap x[i] & x[j]
      Temp = x[i-1];
      x[i-1] = x[j-1];
      x[j-1] = Temp;
//      cout << "x[" << i-1 << "] -> x[" << j-1 << "]" << endl;
      k = N/2;
      while (k < j) {
        j = j-k;
        k = k/2;
      }
    }
    else {
      k = N/2;
      while (k < j) {
        j = j-k;
        k = k/2;
      }
    }
    j = j+k;
  }

//  cout << endl
//    << "**************" << endl
//    << "Performing FFT" << endl
//    << "**************" << endl << endl;

  // Calculate number of stages required for FFT
  // i.e. Log2(N)
  int NumStages=0;
  int Rem = N;
  while (Rem > 1) {
    Rem /=2;
    NumStages +=1;
  }

  // Do FFT on reordered data
  for (int Stage=1;Stage<=NumStages;Stage++) {

    BSep   = int(pow(2,Stage));         // Separation between butterflies
    BWidth = BSep/2;                    // Width of each butterfly
    P      = N/BSep;

    for (int j = 0; j < BWidth; j++) {

      R=P*j;                            // Power of Wn
      TopVal=j;

      // Do all Butterflies in current stage which use same power of Wn
      while (TopVal < N-BWidth) {

        // Bottom input to Radix-2 FFT
        BotVal = TopVal + BWidth;

        // Compute Radix-2 butterfly of TopVal&BotVal
        do_butterfly(TopVal, BotVal, R);

        // Step to next radix-2 butterfly
        TopVal = TopVal + BSep;
      }
    }
  }

//  for (int k = 0; k < N; k++) {
//    cout << "X[" << k << "] = " << x[k].real
//      << " + j " << x[k].imag << endl;
//  }
//  cout << endl
//    << "*********************" << endl
//    << "Calculating Magnitude" << endl
//    << "*********************" << endl << endl;

  for (k = 0; k < N; k++) {
    x[k].real = sqrt((x[k].real*x[k].real) + (x[k].imag*x[k].imag));
  }

//  for (k = 0; k < N; k++) {
//    cout << "|X[" << k << "]| = " << x[k].real << endl;
//  }
}

void entry(int tag, float* data_in) {
  int n;

  for (n = 0; n < N; n++)
  {
    ////////////////
    // Data input //
    ////////////////

    // Read N input Samples
    x[n].real = data_in[n];
    x[n].imag = 0;
  }

//  cout << endl
//    << "*************************************" << endl
//    << "Data Input Complete :"<< n <<" Samples Read." << endl
//    << "*************************************" << endl << endl;
//  for (n = 0; n < N; n++) {
//    cout << "x[" << n << "] = " << x[n].real
//      << " + j " << x[n].imag << endl;
//  }

  ////////////////////////
  // Do FFT calculation //
  ////////////////////////
  do_fft();

  /////////////////
  // Data output //
  /////////////////
//  cout << endl
//    << "***************" << endl
//    << "Outputting Data" << endl
//    << "***************" << endl << endl;


  cout << endl
       << "==========================" << endl
       << "FFT for sample stream # " << tag << endl
       << "==========================" << endl;

  float min = +1e10;
  float max = -1e10;
  for (n = 0; n < N/2+1; n++)
  {
    if (x[n].real > max)
      max = x[n].real;
    if (x[n].real < min)
      min = x[n].real;
  }
  for (n = 0; n < N/2+1; n++)
  {
    int p = (int)((x[n].real - min) / (max - min) * 100.0F);
    for (int i = 0; i < p; i++)
      cout << "*";
    cout << endl;
  }

}


static float* data_in[16];

extern "C"
void initialize_fft(int n_points)
{
  N = n_points;
  for (int i = 0; i < 16; i++)
    data_in[i] = new float[N];

  x = new complex[N];                 // create sample array
  W = new complex[N/2];               // create twiddle array
  for (int k=0;k<N/2;k++) {           // initialise twiddles
     W[k].real=float(cos(2*PI*k/N));
     W[k].imag=float(-1*sin(2*PI*k/N));
  }
}

extern "C"
void transfer_sample(svLogicVecVal* tag, svLogicVecVal* data)
{
  static int n_sample[16] = {0};
  int t;

#ifdef CADENCE
  t = tag->a;
  data_in[t][n_sample[t]++] = (short int)(data->a) / 32768.0F;
#else
  t = tag->aval;
  data_in[t][n_sample[t]++] = (short int)(data->aval) / 32768.0F;
#endif

  if (n_sample[t] >= N)
  {
    entry(t, data_in[t]);
    n_sample[t] = 0;
  }
}
