
// Filename: top.sv

//----------------------------------------------------------------------
//  Copyright (c) 2010 by Doulos.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//----------------------------------------------------------------------

`include "ovm_macros.svh"

package my_pkg;

  `ifdef INCA 
    `include "ovm.svh"
  `else
    import ovm_pkg::*;
  `endif

  class my_tx extends ovm_sequence_item;
  
    typedef enum byte { READ, WRITE } cmd_t;  
  
    rand cmd_t cmd;
    rand int   addr;
    rand int   data;
    
    function new (string name = "");
      super.new(name);
    endfunction: new
  
    constraint c_addr { addr >= 0; addr < 256; }
    constraint c_data { data >= 0; data < 256; }
    
    `ovm_object_utils_begin(my_tx)
      `ovm_field_enum(cmd_t, cmd,  OVM_ALL_ON + OVM_ENUM)
      `ovm_field_int(addr, OVM_ALL_ON + OVM_DEC)
      `ovm_field_int(data, OVM_ALL_ON + OVM_BIN)
    `ovm_object_utils_end
  
  endclass: my_tx
  
  
  class my_sequencer extends ovm_sequencer #(my_tx);

    `ovm_sequencer_utils(my_sequencer)

    function new (string name = "my_sequencer", ovm_component parent);
      super.new(name, parent);
      `ovm_update_sequence_lib_and_item(my_tx)
    endfunction : new

  endclass : my_sequencer


  import "DPI-SC"  context function void   entry(input int id, input byte cmd, input int addr, inout int data);
  import "DPI-SC"  function string get_path(input int id);
  export "DPI-SC"  function tx_done;

  int path_to_id[string];
  event done_ev[5];
    
  function void tx_done(input int id, input byte cmd, input int addr, input int data);
    $display("             vvvvv Called tx_done at %4d, id = %1d, cmd = %1d, addr = %3d, data = %3d",
             $time, id, cmd, addr, data);
    ->done_ev[id];
  endfunction

    
  class my_driver extends ovm_driver #(my_tx);

    `ovm_component_utils(my_driver)
  
    function new(string name, ovm_component parent);
      super.new(name, parent);
    endfunction: new
    
    virtual function void end_of_elaboration;
      `ifdef INCA
        // Hack because Cadence will not return type string from a DPI function!!!!!
        path_to_id["top.m_hier.m_scwrap_1"] = 1;
        path_to_id["top.m_hier.m_scwrap_2"] = 2;
      `else
        path_to_id[get_path(1)] = 1;
        path_to_id[get_path(2)] = 2;
      `endif
    endfunction
    
    virtual task run;

      forever
      begin
        my_tx tx;
        int id;
        #10
        seq_item_port.get_next_item(tx);
        
        ovm_report_info("", $psprintf("OVM TB driving cmd = %1d, addr = %3d, data = %3d}",
                                        tx.cmd, tx.addr, tx.data));

        if ($urandom_range(0,1))
          `ifdef INCA
            id = path_to_id["top.m_hier.m_scwrap_1"];
          `else
            id = path_to_id["/top/m_hier/m_scwrap_1"];
          `endif
        else
          `ifdef INCA
            id = path_to_id["top.m_hier.m_scwrap_2"];
          `else
            id = path_to_id["/top/m_hier/m_scwrap_2"];
          `endif
        
        entry(id, tx.cmd, tx.addr, tx.data); 
      
        // Wait until tx done to avoid pipelining
        //@(done_ev[id]);
        //$display("OVM TB tx done");
       
        seq_item_port.item_done();
      end
    endtask: run
    
  endclass: my_driver


  class my_env extends ovm_env;

    `ovm_component_utils(my_env)
    
    my_sequencer                          m_sequencer;
    my_driver                             m_driver;
  
    function new(string name = "my_env", ovm_component parent = null);
      super.new(name, parent);
    endfunction: new
    
    virtual function void build;
      super.build();
      m_driver    = my_driver   ::type_id::create("m_driver"   , this);
      m_sequencer = my_sequencer::type_id::create("m_sequencer", this);
    endfunction: build
    
    virtual function void connect;
      m_driver.seq_item_port.connect(m_sequencer.seq_item_export);
    endfunction: connect

  endclass: my_env
endpackage: my_pkg
  

module hier;

scwrap m_scwrap_1();
scwrap m_scwrap_2();

endmodule

  
module top;

`ifdef INCA 
  `include "ovm.svh"
`else
  import ovm_pkg::*;
`endif

import my_pkg::*;

hier m_hier();

class my_test extends ovm_test;
  `ovm_component_utils(my_test)
    
  function new(string name, ovm_component parent);
    super.new(name,parent);
  endfunction: new

  my_env m_env;   
 
  virtual function void build;
    super.build();
    set_config_int("m_env.m_sequencer", "count", 20);
    m_env = my_env::type_id::create("m_env", this);
    set_global_timeout(1us);
  endfunction: build
  
endclass: my_test

initial
begin
  run_test("my_test");
end

endmodule

`ifdef INCA
// Needed by Cadence only

module scwrap ()
// The foreign attribute string value must be a SystemC value.
(* integer foreign = "SystemC"; *);
// Verilog port names must match port names exactly as they appear in the
// sc_module class in SystemC; they must also match in order, mode, and type.
endmodule

`endif
